# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import logging
import os
import tempfile
import threading
from argparse import ArgumentParser
from contextlib import contextmanager

from pex.commands.command import OutputMixin, try_open_file, try_run_program
from pex.common import safe_mkdir
from pex.dist_metadata import requires_dists
from pex.interpreter_constraints import InterpreterConstraint
from pex.pex import PEX
from pex.result import Ok, Result
from pex.tools.command import PEXCommand
from pex.tools.commands.digraph import DiGraph
from pex.typing import TYPE_CHECKING
from pex.variables import ENV

if TYPE_CHECKING:
    from typing import IO, Iterator, Tuple

logger = logging.getLogger(__name__)


class Graph(OutputMixin, PEXCommand):
    """Generates a dot graph of the dependencies contained in a PEX file."""

    @staticmethod
    def _create_dependency_graph(pex):
        # type: (PEX) -> DiGraph
        graph = DiGraph(
            pex.path(),
            fontsize="14",
            labelloc="t",
            label="Dependency graph of {} for interpreter {} ({})".format(
                pex.path(),
                pex.interpreter.binary,
                InterpreterConstraint.exact_version(pex.interpreter),
            ),
        )
        marker_environment = pex.interpreter.identity.env_markers.as_dict()
        marker_environment["extra"] = ""
        present_dists = frozenset(dist.metadata.project_name for dist in pex.resolve())
        for dist in pex.resolve():
            graph.add_node(
                name=dist.project_name,
                label="{name} {version}".format(name=dist.project_name, version=dist.version),
                URL="https://pypi.org/project/{name}/{version}".format(
                    name=dist.project_name, version=dist.version
                ),
                target="_blank",
            )
            for req in requires_dists(dist):
                if (
                    req.project_name not in present_dists
                    and req.marker
                    and not req.marker.evaluate(environment=marker_environment)
                ):
                    graph.add_node(
                        name=req.name,
                        color="lightgrey",
                        style="filled",
                        tooltip="inactive requirement",
                        URL="https://pypi.org/project/{name}".format(name=req.project_name),
                        target="_blank",
                    )
                graph.add_edge(
                    start=dist.project_name,
                    end=req.name,
                    label="{specifier}{marker}".format(
                        specifier=req.specifier if req.specifier else "",
                        marker="; {}".format(req.marker) if req.marker else "",
                    )
                    if (req.specifier or req.marker)
                    else None,
                    fontsize="10",
                )
        return graph

    @classmethod
    def add_arguments(cls, parser):
        # type: (ArgumentParser) -> None
        cls.add_output_option(parser, entity="dot graph")
        parser.add_argument(
            "-r",
            "--render",
            action="store_true",
            help="Attempt to render the graph.",
        )
        parser.add_argument(
            "-f",
            "--format",
            default="svg",
            help="The format to render the graph in.",
        )
        parser.add_argument(
            "--open",
            action="store_true",
            help="Attempt to open the graph in the system viewer (implies --render).",
        )
        cls.register_global_arguments(parser)

    def _dot(
        self,
        graph,  # type: DiGraph
        render_fp,  # type: IO
    ):
        # type: (...) -> Result
        read_fd, write_fd = os.pipe()

        def emit():
            with os.fdopen(write_fd, "w") as fp:
                graph.emit(fp)

        emit_thread = threading.Thread(name="{} Emitter".format(__name__), target=emit)
        emit_thread.daemon = True
        emit_thread.start()

        try:
            return try_run_program(
                "dot",
                url="https://graphviz.org/",
                error="Failed to render dependency graph for {}.".format(graph.name),
                args=["-T", self.options.format],
                stdin=read_fd,
                stdout=render_fp,
            )
        finally:
            emit_thread.join()

    @contextmanager
    def _output_for_open(self):
        # type: () -> Iterator[Tuple[IO, str]]
        if self.is_stdout(self.options):
            tmpdir = os.path.join(ENV.PEX_ROOT, "tmp")
            safe_mkdir(tmpdir)
            with tempfile.NamedTemporaryFile(
                prefix="{}.".format(__name__),
                suffix=".deps.{}".format(self.options.format),
                dir=tmpdir,
                delete=False,
            ) as tmp_out:
                yield tmp_out, tmp_out.name
                return

        with self.output(self.options, binary=True) as out:
            yield out, out.name

    def run(self, pex):
        # type: (PEX) -> Result
        graph = self._create_dependency_graph(pex)
        if not (self.options.render or self.options.open):
            with self.output(self.options) as out:
                graph.emit(out)
            return Ok()

        if not self.options.open:
            with self.output(self.options, binary=True) as out:
                return self._dot(graph, out)

        with self._output_for_open() as (out, open_path):
            result = self._dot(graph, out)
            if result.is_error:
                return result

        return try_open_file(
            open_path,
            error="Failed to open dependency graph of {} rendered in {} for viewing.".format(
                pex.path(), open_path
            ),
        )



# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import os
from argparse import ArgumentParser, Namespace

from pex import pex_bootstrapper
from pex.commands.command import GlobalConfigurationError, Main
from pex.compatibility import commonpath
from pex.pex import PEX
from pex.pex_bootstrapper import InterpreterTest
from pex.pex_info import PexInfo
from pex.result import Result, catch
from pex.tools import commands
from pex.tools.command import PEXCommand
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Callable, Optional, Union

    CommandFunc = Callable[[PEX, Namespace], Result]


def simplify_pex_path(pex_path):
    # type: (str) -> str
    # Generate the most concise path possible that is still cut/paste-able to the command line.
    pex_path = os.path.abspath(pex_path)
    cwd = os.getcwd()
    if commonpath((pex_path, cwd)) == cwd:
        pex_path = os.path.relpath(pex_path, cwd)
        # Handle users that do not have . as a PATH entry.
        if not os.path.dirname(pex_path) and os.curdir not in os.environ.get("PATH", "").split(
            os.pathsep
        ):
            pex_path = os.path.join(os.curdir, pex_path)
    return pex_path


class PexTools(Main[PEXCommand]):
    def __init__(self, pex=None):
        # type: (Optional[PEX]) -> None

        pex_prog_path = simplify_pex_path(pex.path()) if pex else None

        # By default, let argparse derive prog from sys.argv[0].
        prog = None  # type: Optional[str]
        if pex:
            prog = "PEX_TOOLS=1 {pex_path}".format(pex_path=pex_prog_path)

        description = "Tools for working with {}.".format(pex_prog_path if pex else "PEX files")
        subparsers_description = (
            "{} can be operated on using any of the following subcommands.".format(
                "The PEX file {}".format(pex_prog_path) if pex else "A PEX file"
            )
        )

        super(PexTools, self).__init__(
            description=description,
            subparsers_description=subparsers_description,
            command_types=commands.all_commands(),
            prog=prog,
        )
        self._pex = pex

    def add_arguments(self, parser):
        # type: (ArgumentParser) -> None
        if self._pex is None:
            parser.add_argument(
                "pex", nargs=1, metavar="PATH", help="The path of the PEX file to operate on."
            )


def main(pex=None):
    # type: (Optional[PEX]) -> Union[int, str]

    pex_tools = PexTools(pex=pex)
    try:
        with pex_tools.parsed_command() as pex_command, TRACER.timed(
            "Executing PEX_TOOLS {}".format(pex_command.name())
        ):
            if pex is None:
                pex_file_path = pex_command.options.pex[0]
                pex_info = PexInfo.from_pex(pex_file_path)
                pex_info.update(PexInfo.from_env())
                interpreter = pex_bootstrapper.find_compatible_interpreter(
                    interpreter_test=InterpreterTest(entry_point=pex_file_path, pex_info=pex_info)
                )
                pex = PEX(pex_file_path, interpreter=interpreter)

            result = catch(pex_command.run, pex)
            result.maybe_display()
            return result.exit_code
    except GlobalConfigurationError as e:
        return str(e)



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.dist_metadata import Distribution
from pex.pep_503 import ProjectName
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class FingerprintedDistribution(object):
    distribution = attr.ib()  # type: Distribution
    fingerprint = attr.ib()  # type: str

    @property
    def location(self):
        # type: () -> str
        return self.distribution.location

    @property
    def project_name(self):
        # type: () -> ProjectName
        return self.distribution.metadata.project_name



# Copyright 2019 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os

from pex.dist_metadata import Requirement
from pex.interpreter import PythonInterpreter, calculate_binary_name
from pex.orderedset import OrderedSet
from pex.pep_425 import CompatibilityTags
from pex.pep_508 import MarkerEnvironment
from pex.platforms import Platform
from pex.result import Error
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Iterable, Iterator, Optional, Tuple, Union

    import attr  # vendor:skip
else:
    from pex.third_party import attr


class RequiresPythonError(Exception):
    """Indicates the impossibility of evaluating Requires-Python metadata."""


@attr.s(frozen=True)
class Target(object):
    id = attr.ib()  # type: str
    platform = attr.ib()  # type: Platform
    marker_environment = attr.ib()  # type: MarkerEnvironment

    def binary_name(self, version_components=2):
        # type: (int) -> str
        return calculate_binary_name(
            platform_python_implementation=self.marker_environment.platform_python_implementation,
            python_version=self.python_version[:version_components]
            if self.python_version and version_components > 0
            else None,
        )

    @property
    def python_version(self):
        # type: () -> Optional[Tuple[int, int]]
        python_version = self.marker_environment.python_version
        return (
            cast("Tuple[int, int]", tuple(map(int, python_version.split(".")))[:2])
            if python_version
            else None
        )

    @property
    def supported_tags(self):
        # type: () -> CompatibilityTags
        raise NotImplementedError()

    @property
    def is_foreign(self):
        # type: () -> bool
        """Does the distribution target represent a foreign platform.

        A foreign platform is one not matching the current interpreter.
        """
        return self.platform not in self.get_interpreter().supported_platforms

    @property
    def python_version_str(self):
        # type: () -> Optional[str]
        return self.marker_environment.python_full_version or self.marker_environment.python_version

    def get_interpreter(self):
        # type: () -> PythonInterpreter
        return PythonInterpreter.get()

    def requires_python_applies(
        self,
        requires_python,  # type: SpecifierSet
        source,  # type: Requirement
    ):
        # type: (...) -> bool
        """Determines if the given python requirement applies to this target.

        :param requires_python: The Python requirement to evaluate.
        :param source: The source of the Python requirement restriction.
        :returns: `True` if the Python requirement applies.
        """

        if not self.python_version_str:
            raise RequiresPythonError(
                "Encountered `Requires-Python: {requires_python}` when evaluating {source} "
                "for applicability but the Python version information needed to evaluate this "
                "requirement is not contained in the target being evaluated for: {target}".format(
                    requires_python=requires_python, source=source, target=self
                )
            )

        # N.B.: The `python_version_str` will be of the form `X.Y` for traditional
        # AbbreviatedPlatform targets with a PYVER of the form `XY`. The Requires-Python metadata
        # (see: https://www.python.org/dev/peps/pep-0345/#requires-python) can contain full versions
        # in its version specifier like `>=3.8.1`. PEP-440 (
        # https://www.python.org/dev/peps/pep-0440/#version-specifiers) specifier missing version
        # components are padded with zeros for all comparison operators besides `===` which can
        # silently lead to incorrect results. If the target platform has a PYVER of 38 we don't know
        # if that platform represents a final target of 3.8.0 or 3.8.1 or some other 3.8 version,
        # but Pip, which follows PEP-440, will evaluate `>=3.8.1` as false and exclude the
        # distribution from consideration.
        #
        # Since our evaluation of `requires_python_applies` (and `requirement_applies`) is always
        # upon _results_ of an underlying Pip resolve (when we resolve from a PEX repository or a
        # lock file), the damage is already done when the distribution has already been incorrectly
        # excluded as in the example above; so we will not have to evaluate it. The other case is
        # when the distribution was incorrectly included. It's this case we need to contend with.
        #
        # If we use the same logic as Pip for an abbreviated platform resolve against a PEX or
        # lockfile, we'll include too much in the resulting PEX or lockfile. However, when the
        # resulting PEX or lockfile gets used "for real" there will be a local interpreter involved
        # and the `Requires-Python` will be correctly evaluated leading to the distribution not
        # being activated. This result is correct at the expense of wasting the space needed to
        # carry along the distribution that was not activated.
        #
        # It turns out that by just passing "X.Y" to SpecifierSet.__contains__ does evaluate "X.Y"
        # as if it were "X.Y.0" when zero-padding is appropriate (since it implements PEP-440); so
        # we get the Pip emulating behavior described above.
        return self.python_version_str in requires_python

    def requirement_applies(
        self,
        requirement,  # type: Requirement
        extras=(),  # type: Iterable[str]
    ):
        # type: (...) -> bool
        """Determines if the given requirement applies to this target.

        :param requirement: The requirement to evaluate.
        :param extras: Optional active extras.
        :returns: `True` if the requirement applies.
        """
        if requirement.marker is None:
            return True

        if not extras:
            # Provide an empty extra to safely evaluate the markers without matching any extra.
            extras = ("",)
        for extra in extras:
            environment = self.marker_environment.as_dict()
            environment["extra"] = extra
            if requirement.marker.evaluate(environment=environment):
                return True

        return False

    def __str__(self):
        # type: () -> str
        return str(self.platform.tag)

    def render_description(self):
        raise NotImplementedError()


@attr.s(frozen=True)
class LocalInterpreter(Target):
    @classmethod
    def create(cls, interpreter=None):
        # type: (Optional[PythonInterpreter]) -> LocalInterpreter
        python_interpreter = interpreter or PythonInterpreter.get()
        return cls(
            id=python_interpreter.binary.replace(os.sep, ".").lstrip("."),
            platform=python_interpreter.platform,
            marker_environment=python_interpreter.identity.env_markers,
            interpreter=python_interpreter,
        )

    interpreter = attr.ib()  # type: PythonInterpreter

    def binary_name(self, version_components=2):
        # type: (int) -> str
        return self.interpreter.identity.binary_name(version_components=version_components)

    @property
    def python_version(self):
        # type: () -> Tuple[int, int]
        return self.interpreter.identity.version[:2]

    @property
    def is_foreign(self):
        # type: () -> bool
        return False

    @property
    def python_version_str(self):
        # type: () -> str
        return self.interpreter.identity.version_str

    def get_interpreter(self):
        # type: () -> PythonInterpreter
        return self.interpreter

    @property
    def supported_tags(self):
        return self.interpreter.identity.supported_tags

    def __str__(self):
        # type: () -> str
        return self.interpreter.binary

    def render_description(self):
        return "{platform} interpreter at {path}".format(
            platform=self.interpreter.platform.tag, path=self.interpreter.binary
        )


@attr.s(frozen=True)
class AbbreviatedPlatform(Target):
    @classmethod
    def create(
        cls,
        platform,  # type: Platform
        manylinux=None,  # type: Optional[str]
    ):
        # type: (...) -> AbbreviatedPlatform
        return cls(
            id=str(platform.tag),
            marker_environment=MarkerEnvironment.from_platform(platform),
            platform=platform,
            manylinux=manylinux,
        )

    manylinux = attr.ib()  # type: Optional[str]

    @property
    def supported_tags(self):
        # type: () -> CompatibilityTags
        return self.platform.supported_tags(manylinux=self.manylinux)

    def render_description(self):
        return "abbreviated platform {platform}".format(platform=self.platform.tag)


def current():
    # type: () -> LocalInterpreter
    return LocalInterpreter.create()


@attr.s(frozen=True)
class CompletePlatform(Target):
    @classmethod
    def from_interpreter(cls, interpreter):
        # type: (PythonInterpreter) -> CompletePlatform
        return cls.create(
            marker_environment=interpreter.identity.env_markers,
            supported_tags=interpreter.identity.supported_tags,
        )

    @classmethod
    def create(
        cls,
        marker_environment,  # type: MarkerEnvironment
        supported_tags,  # type: CompatibilityTags
    ):
        # type: (...) -> CompletePlatform

        platform = Platform.from_tag(supported_tags[0])
        return cls(
            id=str(platform.tag),
            marker_environment=marker_environment,
            platform=platform,
            supported_tags=supported_tags,
        )

    _supported_tags = attr.ib()  # type: CompatibilityTags

    @property
    def supported_tags(self):
        # type: () -> CompatibilityTags
        return self._supported_tags

    def render_description(self):
        return "complete platform {platform}".format(platform=self.platform.tag)


@attr.s(frozen=True)
class Targets(object):
    @classmethod
    def from_target(cls, target):
        # type: (Target) -> Targets
        if isinstance(target, AbbreviatedPlatform):
            return cls(platforms=(target.platform,), assume_manylinux=target.manylinux)
        elif isinstance(target, CompletePlatform):
            return cls(complete_platforms=(target,))
        else:
            return cls(interpreters=(target.get_interpreter(),))

    interpreters = attr.ib(default=())  # type: Tuple[PythonInterpreter, ...]
    complete_platforms = attr.ib(default=())  # type: Tuple[CompletePlatform, ...]
    platforms = attr.ib(default=())  # type: Tuple[Optional[Platform], ...]
    assume_manylinux = attr.ib(default=None)  # type: Optional[str]

    @property
    def interpreter(self):
        # type: () -> Optional[PythonInterpreter]
        if not self.interpreters:
            return None
        return PythonInterpreter.latest_release_of_min_compatible_version(self.interpreters)

    def unique_targets(self, only_explicit=False):
        # type: (bool) -> OrderedSet[Target]

        def iter_targets():
            # type: () -> Iterator[Target]
            if (
                not only_explicit
                and not self.interpreters
                and not self.platforms
                and not self.complete_platforms
            ):
                # No specified targets, so just build for the current interpreter (on the current
                # platform).
                yield current()
                return

            for interpreter in self.interpreters:
                # Build for the specified local interpreters (on the current platform).
                yield LocalInterpreter.create(interpreter)

            for platform in self.platforms:
                if platform is None and not self.interpreters:
                    # Build for the current platform (None) only if not done already (ie: no
                    # interpreters were specified).
                    yield current()
                elif platform is not None:
                    # Build for specific platforms.
                    yield AbbreviatedPlatform.create(platform, manylinux=self.assume_manylinux)

            for complete_platform in self.complete_platforms:
                yield complete_platform

        return OrderedSet(iter_targets())

    def require_unique_target(self, purpose):
        # type: (str) -> Union[Target, Error]
        resolved_targets = self.unique_targets()
        if len(resolved_targets) != 1:
            return Error(
                "A single target is required for {purpose}.\n"
                "There were {count} targets selected:\n"
                "{targets}".format(
                    purpose=purpose,
                    count=len(resolved_targets),
                    targets="\n".join(
                        "{index}. {target}".format(index=index, target=target)
                        for index, target in enumerate(resolved_targets, start=1)
                    ),
                )
            )
        return cast(Target, next(iter(resolved_targets)))

    def require_at_most_one_target(self, purpose):
        # type: (str) -> Union[Optional[Target], Error]
        resolved_targets = self.unique_targets(only_explicit=False)
        if len(resolved_targets) > 1:
            return Error(
                "At most a single target is required for {purpose}.\n"
                "There were {count} targets selected:\n"
                "{targets}".format(
                    purpose=purpose,
                    count=len(resolved_targets),
                    targets="\n".join(
                        "{index}. {target}".format(index=index, target=target)
                        for index, target in enumerate(resolved_targets, start=1)
                    ),
                )
            )
        try:
            return cast(Target, next(iter(resolved_targets)))
        except StopIteration:
            return None



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
from textwrap import dedent

import pytest

from pex.common import safe_open, temporary_dir
from pex.testing import PY27, ensure_python_venv, make_env, run_pex_command


def test_extras_isolation():
    # type: () -> None
    # Here we ensure one of our extras, `subprocess32`, is properly isolated in the transition from
    # pex bootstrapping where it is imported by `pex.executor` to execution of user code.
    python, pip = ensure_python_venv(PY27)
    subprocess.check_call([pip, "install", "subprocess32"])
    with temporary_dir() as td:
        src_dir = os.path.join(td, "src")
        with safe_open(os.path.join(src_dir, "test_issues_745.py"), "w") as fp:
            fp.write(
                dedent(
                    """\
                    import subprocess32

                    print(subprocess32.__file__)
                    """
                )
            )

        pex_file = os.path.join(td, "test.pex")

        run_pex_command(
            [
                "--sources-directory={}".format(src_dir),
                "--entry-point=test_issues_745",
                "-o",
                pex_file,
            ],
            python=python,
        )

        # The pex runtime should scrub subprocess32 since it comes from site-packages and so we should
        # not have access to it.
        with pytest.raises(subprocess.CalledProcessError):
            subprocess.check_call([python, pex_file])

        # But if the pex has a declared dependency on subprocess32 we should be able to find the
        # subprocess32 bundled into the pex.
        pex_root = os.path.realpath(os.path.join(td, "pex_root"))
        run_pex_command(
            [
                "subprocess32",
                "--sources-directory={}".format(src_dir),
                "--entry-point=test_issues_745",
                "-o",
                pex_file,
            ],
            python=python,
        )

        output = subprocess.check_output([python, pex_file], env=make_env(PEX_ROOT=pex_root))

        subprocess32_location = os.path.realpath(output.decode("utf-8").strip())
        assert subprocess32_location.startswith(pex_root)



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).
import os.path

import pytest

from pex.cli.testing import run_pex3
from pex.compatibility import commonpath
from pex.testing import IntegResults, built_wheel, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Iterator


@pytest.fixture
def wheel():
    # type: () -> Iterator[str]
    with built_wheel(name="not_boto", version="2.49.0a1") as whl:
        yield whl


def test_resolve_arbitrary_equality(
    tmpdir,  # type: Any
    wheel,  # type: str
):
    # type: (...) -> None

    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        "lock",
        "create",
        "--no-pypi",
        "--find-links",
        os.path.dirname(wheel),
        "not_boto===2.49.0a1",
        "--indent",
        "2",
        "-o",
        lock,
    ).assert_success()

    pex_root = os.path.join(str(tmpdir), "pex_root")

    def create_pex_from_lock(*additional_args):
        # type: (*str) -> IntegResults
        return run_pex_command(
            args=[
                "--pex-root",
                pex_root,
                "--runtime-pex-root",
                pex_root,
                "--lock",
                lock,
            ]
            + list(additional_args)
            + ["--", "-c", "import not_boto; print(not_boto.__file__)"]
        )

    result = create_pex_from_lock()
    result.assert_success()
    assert pex_root == commonpath((pex_root, result.output.strip()))

    result = create_pex_from_lock("not_boto===2.49.0a1")
    result.assert_success()
    assert pex_root == commonpath((pex_root, result.output.strip()))

    result = create_pex_from_lock("not_boto===2.49a1")
    result.assert_failure()

    pex_repository = os.path.join(str(tmpdir), "pex_repository")
    run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--lock",
            lock,
            "-o",
            pex_repository,
        ]
    ).assert_success()

    def create_pex_from_pex_repository(requirement):
        # type: (str) -> IntegResults
        return run_pex_command(
            args=[
                "--pex-root",
                pex_root,
                "--runtime-pex-root",
                pex_root,
                "--pex-repository",
                pex_repository,
                requirement,
                "--",
                "-c",
                "import not_boto; print(not_boto.__file__)",
            ]
        )

    result = create_pex_from_pex_repository("not_boto===2.49.0a1")
    result.assert_success()
    assert pex_root == commonpath((pex_root, result.output.strip()))

    result = create_pex_from_pex_repository("not_boto===2.49a1")
    result.assert_failure()



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).
import os.path
import sys

from pex.cli.testing import run_pex3
from pex.compatibility import commonpath
from pex.testing import PY310, ensure_python_interpreter, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_pypy_impl_tag_handling(tmpdir):
    # type: (Any) -> None

    pex_root = os.path.realpath(os.path.join(str(tmpdir), "pex_root"))
    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        "lock",
        "create",
        "--pex-root",
        pex_root,
        "lexid",
        "--style",
        "universal",
        "--resolver-version",
        "pip-2020-resolver",
        "--interpreter-constraint",
        ">=3.7,<4",
        "-o",
        lock,
        "--indent",
        "2",
    ).assert_success()

    python = sys.executable if sys.version_info[:2] >= (3, 7) else ensure_python_interpreter(PY310)
    result = run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--lock",
            lock,
            "--",
            "-c",
            "import lexid; print(lexid.__file__)",
        ],
        python=python,
    )
    result.assert_success()
    assert pex_root == commonpath([pex_root, os.path.realpath(result.output.strip())])



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
from collections import OrderedDict

from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import (
    InterpreterConstraints,
    UnsatisfiableInterpreterConstraintsError,
)
from pex.orderedset import OrderedSet
from pex.pex_bootstrapper import iter_compatible_interpreters, normalize_path
from pex.platforms import Platform
from pex.targets import CompletePlatform, Targets
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING
from pex.variables import ENV

if TYPE_CHECKING:
    from typing import FrozenSet, Iterator, Optional, Set, Tuple

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class InterpreterConfiguration(object):
    _python_path = attr.ib(default=None)  # type: Optional[Tuple[str, ...]]
    pythons = attr.ib(default=())  # type: Tuple[str, ...]
    interpreter_constraints = attr.ib(
        default=InterpreterConstraints()
    )  # type: InterpreterConstraints

    @property
    def python_path(self):
        # type: () -> Optional[Tuple[str, ...]]
        # TODO(#1075): stop looking at PEX_PYTHON_PATH and solely consult the `--python-path` flag.
        return self._python_path or ENV.PEX_PYTHON_PATH

    def iter_interpreters(self):
        # type: () -> Iterator[PythonInterpreter]

        if self.pythons:
            with TRACER.timed("Resolving interpreters", V=2):

                def to_python_interpreter(full_path_or_basename):
                    if os.path.isfile(full_path_or_basename):
                        return PythonInterpreter.from_binary(full_path_or_basename)
                    else:
                        interpreter = PythonInterpreter.from_env(
                            full_path_or_basename, paths=normalize_path(self.python_path)
                        )
                        if interpreter is None:
                            raise InterpreterNotFound(
                                "Failed to find interpreter: {}".format(full_path_or_basename)
                            )
                        return interpreter

                for python in self.pythons:
                    yield to_python_interpreter(python)

        if self.interpreter_constraints:
            with TRACER.timed("Resolving interpreters", V=2):
                try:
                    for interp in iter_compatible_interpreters(
                        path=self.python_path,
                        interpreter_constraints=self.interpreter_constraints,
                    ):
                        yield interp
                except UnsatisfiableInterpreterConstraintsError as e:
                    raise InterpreterConstraintsNotSatisfied(
                        e.create_message("Could not find a compatible interpreter.")
                    )

    def resolve_interpreters(self):
        # type: () -> OrderedSet[PythonInterpreter]
        """Resolves the interpreters satisfying the interpreter configuration.

        :raise: :class:`InterpreterNotFound` specific --python interpreters were requested but could
            not be found.
        :raise: :class:`InterpreterConstraintsNotSatisfied` if --interpreter-constraint were
            specified but no conforming interpreters could be found.
        """
        return OrderedSet(self.iter_interpreters())


class TargetConfigurationError(Exception):
    """Indicates a problem configuring resolve targets."""


class InterpreterNotFound(TargetConfigurationError):
    """Indicates an explicitly requested interpreter could not be found."""


class InterpreterConstraintsNotSatisfied(TargetConfigurationError):
    """Indicates no interpreter meeting the requested constraints could be found."""


def _interpreter_compatible_platforms(
    requested_complete_platforms,  # type: OrderedSet[CompletePlatform]
    candidate_interpreter,  # type: PythonInterpreter
):
    # type: (...) -> FrozenSet[CompletePlatform]
    compatible_complete_platforms = []
    for requested_complete in requested_complete_platforms:
        # if there was an explicit complete platform specified, only use the local interpreter
        # when the interpreter's tags are a subset of the complete platform's: tags supported by
        # the interpreter but not the complete platform may result in incompatible wheels being
        # chosen, if the interpreter was used directly
        interpreter_platform = CompletePlatform.from_interpreter(candidate_interpreter)
        missing_tags = set(interpreter_platform.supported_tags) - set(
            requested_complete.supported_tags
        )
        if missing_tags:
            TRACER.log(
                "Rejected candidate interpreter {} for complete platform {} since interpreter supports {} extra tags".format(
                    candidate_interpreter,
                    requested_complete.platform,
                    len(missing_tags),
                ),
                V=3,
            )
            TRACER.log(
                "Extra tags supported by {} but not supported by requested complete platform {}: {}".format(
                    candidate_interpreter,
                    requested_complete.platform,
                    ", ".join(map(str, missing_tags)),
                ),
                V=9,
            )
        else:
            TRACER.log(
                "Accepted resolution of {} for complete platform {}".format(
                    candidate_interpreter,
                    requested_complete,
                ),
                V=3,
            )
            compatible_complete_platforms.append(requested_complete)

    return frozenset(compatible_complete_platforms)


@attr.s(frozen=True)
class TargetConfiguration(object):
    interpreter_configuration = attr.ib(
        default=InterpreterConfiguration()
    )  # type: InterpreterConfiguration

    @property
    def pythons(self):
        # type: () -> Tuple[str, ...]
        return self.interpreter_configuration.pythons

    @property
    def interpreter_constraints(self):
        # type: () -> InterpreterConstraints
        return self.interpreter_configuration.interpreter_constraints

    complete_platforms = attr.ib(default=())  # type: Tuple[CompletePlatform, ...]
    platforms = attr.ib(default=())  # type: Tuple[Optional[Platform], ...]
    assume_manylinux = attr.ib(default="manylinux2014")  # type: Optional[str]
    resolve_local_platforms = attr.ib(default=False)  # type: bool

    def resolve_targets(self):
        # type: () -> Targets
        """Resolves the targets satisfying the target configuration.

        :raise: :class:`InterpreterNotFound` specific --python interpreters were requested but could
            not be found.
        :raise: :class:`InterpreterConstraintsNotSatisfied` if --interpreter-constraint were
            specified but no conforming interpreters could be found.
        """
        interpreters = self.interpreter_configuration.resolve_interpreters()

        requested_platforms = OrderedSet(self.platforms)  # type: OrderedSet[Optional[Platform]]
        requested_complete_platforms = OrderedSet(
            self.complete_platforms
        )  # type: OrderedSet[CompletePlatform]
        if (requested_platforms or requested_complete_platforms) and self.resolve_local_platforms:
            # If any platform or complete_platform matches a local interpreter, we remove that
            # platform or complete_platform from the requested_* set and instead use the
            # interpreter.

            platform_strs = list(map(str, requested_complete_platforms)) + list(
                map(str, requested_platforms)
            )
            with TRACER.timed(
                "Searching for local interpreters matching {}".format(", ".join(platform_strs))
            ):
                candidate_interpreters = OrderedSet(
                    iter_compatible_interpreters(path=self.interpreter_configuration.python_path)
                )  # type: OrderedSet[PythonInterpreter]
                candidate_interpreters.add(PythonInterpreter.get())
                for candidate_interpreter in candidate_interpreters:
                    resolved_platforms = candidate_interpreter.supported_platforms.intersection(
                        requested_platforms
                    )
                    if resolved_platforms:
                        for resolved_platform in resolved_platforms:
                            TRACER.log(
                                "Resolved {} for platform {}".format(
                                    candidate_interpreter, resolved_platform
                                )
                            )
                            requested_platforms.remove(resolved_platform)
                        interpreters.add(candidate_interpreter)

                    resolved_complete_platforms = _interpreter_compatible_platforms(
                        requested_complete_platforms, candidate_interpreter
                    )
                    if resolved_complete_platforms:
                        for resolved_complete_platform in resolved_complete_platforms:
                            TRACER.log(
                                "Resolved {} for complete platform {}".format(
                                    candidate_interpreter, resolved_complete_platform
                                )
                            )
                            requested_complete_platforms.remove(resolved_complete_platform)
                        interpreters.add(candidate_interpreter)

            if requested_platforms or requested_complete_platforms:
                platform_strs = list(map(str, requested_complete_platforms)) + list(
                    map(str, requested_platforms)
                )
                TRACER.log(
                    "Could not resolve a local interpreter for {}, will resolve only binary "
                    "distributions for {}.".format(
                        ", ".join(platform_strs),
                        "this platform" if len(platform_strs) == 1 else "these platforms",
                    )
                )

        return Targets(
            interpreters=tuple(interpreters),
            complete_platforms=tuple(requested_complete_platforms),
            platforms=tuple(requested_platforms),
            assume_manylinux=self.assume_manylinux,
        )



from __future__ import absolute_import, division, unicode_literals

from xml.sax.xmlreader import AttributesNSImpl

from ..constants import adjustForeignAttributes, unadjustForeignAttributes

prefix_mapping = {}
for prefix, localName, namespace in adjustForeignAttributes.values():
    if prefix is not None:
        prefix_mapping[prefix] = namespace


def to_sax(walker, handler):
    """Call SAX-like content handler based on treewalker walker

    :arg walker: the treewalker to use to walk the tree to convert it

    :arg handler: SAX handler to use

    """
    handler.startDocument()
    for prefix, namespace in prefix_mapping.items():
        handler.startPrefixMapping(prefix, namespace)

    for token in walker:
        type = token["type"]
        if type == "Doctype":
            continue
        elif type in ("StartTag", "EmptyTag"):
            attrs = AttributesNSImpl(token["data"],
                                     unadjustForeignAttributes)
            handler.startElementNS((token["namespace"], token["name"]),
                                   token["name"],
                                   attrs)
            if type == "EmptyTag":
                handler.endElementNS((token["namespace"], token["name"]),
                                     token["name"])
        elif type == "EndTag":
            handler.endElementNS((token["namespace"], token["name"]),
                                 token["name"])
        elif type in ("Characters", "SpaceCharacters"):
            handler.characters(token["data"])
        elif type == "Comment":
            pass
        else:
            assert False, "Unknown token type"

    for prefix, namespace in prefix_mapping.items():
        handler.endPrefixMapping(prefix)
    handler.endDocument()



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from abc import abstractmethod

from pex.dist_metadata import Distribution, Requirement
from pex.fingerprinted_distribution import FingerprintedDistribution
from pex.pip.version import PipVersionValue
from pex.resolve.lockfile.model import Lockfile
from pex.sorted_tuple import SortedTuple
from pex.targets import Target, Targets
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, Optional, Tuple

    import attr  # vendor:skip
else:
    from pex.third_party import attr


# Derived from notes in the bandersnatch PyPI mirroring tool:
# https://github.com/pypa/bandersnatch/blob/1485712d6aa77fba54bbf5a2df0d7314124ad097/src/bandersnatch/default.conf#L30-L35
MAX_PARALLEL_DOWNLOADS = 10


class ResolveError(Exception):
    """Indicates an error resolving requirements for a PEX."""


class Untranslatable(ResolveError):
    pass


class Unsatisfiable(ResolveError):
    pass


def _sorted_requirements(requirements):
    # type: (Optional[Iterable[Requirement]]) -> SortedTuple[Requirement]
    return SortedTuple(requirements, key=lambda req: str(req)) if requirements else SortedTuple()


@attr.s(frozen=True)
class InstalledDistribution(object):
    """A distribution target, and the installed distribution that satisfies it.

    If installed distribution directly satisfies a user-specified requirement, that requirement is
    included.
    """

    target = attr.ib()  # type: Target
    fingerprinted_distribution = attr.ib()  # type: FingerprintedDistribution
    direct_requirements = attr.ib(
        converter=_sorted_requirements, factory=SortedTuple
    )  # type: SortedTuple[Requirement]

    @property
    def distribution(self):
        # type: () -> Distribution
        return self.fingerprinted_distribution.distribution

    @property
    def fingerprint(self):
        # type: () -> str
        return self.fingerprinted_distribution.fingerprint

    def with_direct_requirements(self, direct_requirements=None):
        # type: (Optional[Iterable[Requirement]]) -> InstalledDistribution
        direct_requirements = _sorted_requirements(direct_requirements)
        if direct_requirements == self.direct_requirements:
            return self
        return InstalledDistribution(
            self.target,
            self.fingerprinted_distribution,
            direct_requirements=direct_requirements,
        )


@attr.s(frozen=True)
class Installed(object):
    installed_distributions = attr.ib()  # type: Tuple[InstalledDistribution, ...]


class Resolver(object):
    @abstractmethod
    def is_default_repos(self):
        # type: () -> bool
        raise NotImplementedError()

    @abstractmethod
    def resolve_lock(
        self,
        lock,  # type: Lockfile
        targets=Targets(),  # type: Targets
        pip_version=None,  # type: Optional[PipVersionValue]
    ):
        # type: (...) -> Installed
        raise NotImplementedError()

    def resolve_requirements(
        self,
        requirements,  # type: Iterable[str]
        targets=Targets(),  # type: Targets
        pip_version=None,  # type: Optional[PipVersionValue]
    ):
        # type: (...) -> Installed
        raise NotImplementedError()



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import itertools
import os.path

import pytest

from pex.cli.testing import run_pex3
from pex.compatibility import urlparse
from pex.interpreter import PythonInterpreter
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.resolve.locked_resolve import Artifact, FileArtifact, LockStyle, TargetSystem
from pex.resolve.lockfile import json_codec
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.resolved_requirement import Pin
from pex.sorted_tuple import SortedTuple
from pex.testing import IntegResults, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Iterable, List, Set


@pytest.mark.parametrize(
    "style",
    [str(lock_style) for lock_style in LockStyle.values() if lock_style is not LockStyle.UNIVERSAL],
)
def test_non_universal_target_system_unsupported(style):
    # type: (str) -> None

    result = run_pex3("lock", "create", "--style", style, "--target-system", "linux", "ansicolors")
    result.assert_failure()
    assert (
        "The --target-system option only applies to --style {universal} locks.\n".format(
            universal=LockStyle.UNIVERSAL
        )
        == result.error
    )


def run_lock(
    lock_file,  # type: str
    extra_args,  # type: Iterable[str]
    *target_systems  # type: TargetSystem.Value
):
    # type: (...) -> IntegResults
    args = ["lock", "create", "--style", "universal", "-o", lock_file, "--indent", "2"]
    for target_system in target_systems:
        args.extend(("--target-system", str(target_system)))
    args.extend(extra_args)
    return run_pex3(*args)


def lock(
    tmpdir,  # type: Any
    extra_args,  # type: Iterable[str]
    *target_systems  # type: TargetSystem.Value
):
    # type: (...) -> Lockfile

    lock_file = os.path.join(
        str(tmpdir),
        "lock{id}.json".format(
            id="-{}".format("-".join(map(str, target_systems))) if target_systems else ""
        ),
    )
    run_lock(lock_file, extra_args, *target_systems).assert_success()
    return json_codec.load(lock_file)


def assert_file_artifact(artifact):
    # type: (Artifact) -> FileArtifact
    assert isinstance(artifact, FileArtifact)
    return artifact


def assert_wheel(artifact):
    # type: (Artifact) -> FileArtifact
    assert urlparse.urlparse(artifact.url).path.endswith(".whl")
    return assert_file_artifact(artifact)


def test_target_system_universal(tmpdir):
    # type: (Any) -> None

    target_systems_powerset = sorted(
        itertools.chain.from_iterable(
            itertools.combinations(TargetSystem.values(), radix)
            for radix in range(len(TargetSystem.values()) + 1)
        )
    )
    locks = [
        lock(tmpdir, ["ansicolors==1.1.8"], *target_systems)
        for target_systems in target_systems_powerset
    ]
    assert len(locks) == len(target_systems_powerset)

    unique_resolves = set(lockfile.locked_resolves for lockfile in locks)
    assert 1 == len(unique_resolves), "Expected all resolves to be identical."

    assert 1 == len(locks[0].locked_resolves)
    locked_resolve = locks[0].locked_resolves[0]

    assert 1 == len(locked_resolve.locked_requirements)
    locked_requirement = locked_resolve.locked_requirements[0]

    assert Pin(ProjectName("ansicolors"), Version("1.1.8")) == locked_requirement.pin
    assert_wheel(locked_requirement.artifact)

    assert 1 == len(
        locked_requirement.additional_artifacts
    ), "Expected one additional source artifact."
    additional_artifact = locked_requirement.additional_artifacts[0]
    assert assert_file_artifact(additional_artifact).is_source


def test_target_system_platform_specific(
    tmpdir,  # type: Any
    py310,  # type: PythonInterpreter
):
    # type: (...) -> None

    # The lineup of psutil 5.9.1 artifacts compatible with CPython==3.10.* out on PyPI is:
    # psutil-5.9.1.tar.gz
    # psutil-5.9.1-cp310-cp310-win_amd64.whl
    # psutil-5.9.1-cp310-cp310-win32.whl
    # psutil-5.9.1-cp310-cp310-manylinux_2_12_x86_64.manylinux2010_x86_64.manylinux_2_17_x86_64.manylinux2014_x86_64.whl
    # psutil-5.9.1-cp310-cp310-manylinux_2_12_i686.manylinux2010_i686.manylinux_2_17_i686.manylinux2014_i686.whl
    # psutil-5.9.1-cp310-cp310-macosx_10_9_x86_64.whl
    #
    # There are additional artifacts for CPython 2.7 and 3.{6,7,8,9}

    def assert_expected_artifacts(
        expected_wheel_count,  # type: int
        *target_systems  # type: TargetSystem.Value
    ):
        # type: (...) -> Set[FileArtifact]

        lockfile = lock(
            tmpdir,
            [
                "--python-path",
                py310.binary,
                "--interpreter-constraint",
                "CPython==3.10.*",
                "psutil==5.9.1",
            ],
            *target_systems
        )

        assert 1 == len(lockfile.locked_resolves)
        locked_resolve = lockfile.locked_resolves[0]

        assert 1 == len(locked_resolve.locked_requirements)
        locked_requirement = locked_resolve.locked_requirements[0]

        assert Pin(ProjectName("psutil"), Version("5.9.1")) == locked_requirement.pin

        wheel_artifacts = set()  # type: Set[FileArtifact]
        sdist_artifacts = []  # type: List[FileArtifact]
        wheel_artifacts.add(assert_wheel(locked_requirement.artifact))
        for additional_artifact in locked_requirement.additional_artifacts:
            file_artifact = assert_file_artifact(additional_artifact)
            if file_artifact.is_source:
                sdist_artifacts.append(file_artifact)
            else:
                wheel_artifacts.add(assert_wheel(file_artifact))
        assert 1 == len(sdist_artifacts), "Expected one additional artifact to be the sdist."
        assert expected_wheel_count == len(wheel_artifacts)
        return wheel_artifacts

    linux_wheels = assert_expected_artifacts(2, TargetSystem.LINUX)
    mac_wheels = assert_expected_artifacts(1, TargetSystem.MAC)
    windows_wheels = assert_expected_artifacts(2, TargetSystem.WINDOWS)

    linux_and_mac_wheels = assert_expected_artifacts(3, TargetSystem.LINUX, TargetSystem.MAC)
    assert linux_wheels | mac_wheels == linux_and_mac_wheels

    linux_and_mac_and_windows_wheels = assert_expected_artifacts(
        5, TargetSystem.LINUX, TargetSystem.MAC, TargetSystem.WINDOWS
    )
    assert linux_wheels | mac_wheels | windows_wheels == linux_and_mac_and_windows_wheels

    assert linux_and_mac_and_windows_wheels == assert_expected_artifacts(5)


def test_issue_1821(
    tmpdir,  # type: Any
    py310,  # type: PythonInterpreter
):
    # type: (...) -> None

    target_options = [
        "--python-path",
        py310.binary,
        "--interpreter-constraint",
        "CPython>=3.10,<4",
    ]

    args = target_options + [
        "--style",
        "universal",
        "--resolver-version",
        "pip-2020-resolver",
        "cryptography==36.0.2",
        "docker==5.0.3",
    ]

    # This lock solution requires pywin32 227 for windows, but that is a wheel-only distribution
    # with no wheels published for CPython 3.10 (see: https://pypi.org/project/pywin32/227/#files);
    # so, without restricting the target operating systems for the lock, we expect failure.
    result = run_lock(os.path.join(str(tmpdir), "lock"), args)
    result.assert_failure()
    assert (
        "ERROR: Could not find a version that satisfies the requirement "
        'pywin32==227; sys_platform == "win32" (from docker)'
    ) in result.error
    assert (
        'ERROR: No matching distribution found for pywin32==227; sys_platform == "win32"'
    ) in result.error

    lockfile = lock(tmpdir, args, TargetSystem.LINUX, TargetSystem.MAC)
    assert SortedTuple((TargetSystem.LINUX, TargetSystem.MAC)) == lockfile.target_systems
    assert lockfile.source is not None
    run_pex_command(
        args=["--lock", lockfile.source, "--", "-c", "import docker"], python=py310.binary
    ).assert_success()

    # Check that lock updates respect target systems.
    update_args = ["lock", "update"] + target_options + ["--dry-run", "check", lockfile.source]
    run_pex3(*update_args).assert_success()



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import ast
import os
import sys
from site import USER_SITE
from types import ModuleType

from pex import third_party
from pex.bootstrap import Bootstrap
from pex.common import die
from pex.dist_metadata import CallableEntryPoint, Distribution, EntryPoint, find_distributions
from pex.environment import PEXEnvironment
from pex.executor import Executor
from pex.finders import get_entry_point_from_console_script, get_script_from_distributions
from pex.inherit_path import InheritPath
from pex.interpreter import PythonInterpreter
from pex.layout import Layout
from pex.orderedset import OrderedSet
from pex.pex_info import PexInfo
from pex.targets import LocalInterpreter
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING
from pex.util import named_temporary_file
from pex.variables import ENV, Variables

if TYPE_CHECKING:
    from typing import (
        Any,
        Dict,
        Iterable,
        Iterator,
        List,
        Mapping,
        NoReturn,
        Optional,
        Set,
        Tuple,
        TypeVar,
        Union,
    )

    _K = TypeVar("_K")
    _V = TypeVar("_V")


class IsolatedSysPath(object):
    @staticmethod
    def _expand_paths(*paths):
        # type: (*str) -> Iterator[str]
        for path in paths:
            yield path
            if not os.path.isabs(path):
                yield os.path.abspath(path)
            yield os.path.realpath(path)

    @classmethod
    def for_pex(
        cls,
        interpreter,  # type: PythonInterpreter
        pex,  # type: str
        pex_pex=None,  # type: Optional[str]
    ):
        # type: (...) -> IsolatedSysPath
        sys_path = OrderedSet(interpreter.sys_path)
        sys_path.add(pex)
        sys_path.add(Bootstrap.locate().path)
        if pex_pex:
            sys_path.add(pex_pex)

        site_packages = OrderedSet()  # type: OrderedSet[str]
        for site_lib in interpreter.site_packages:
            TRACER.log("Discarding site packages path: {site_lib}".format(site_lib=site_lib))
            site_packages.add(site_lib)

        extras_paths = OrderedSet()  # type: OrderedSet[str]
        for extras_path in interpreter.extras_paths:
            TRACER.log("Discarding site extras path: {extras_path}".format(extras_path=extras_path))
            extras_paths.add(extras_path)

        return cls(
            sys_path=sys_path,
            site_packages=site_packages,
            extras_paths=extras_paths,
            is_venv=interpreter.is_venv,
        )

    def __init__(
        self,
        sys_path,  # type: Iterable[str]
        site_packages,  # type: Iterable[str]
        extras_paths=(),  # type: Iterable[str]
        is_venv=False,  # type: bool
    ):
        # type: (...) -> None
        self._sys_path_entries = tuple(self._expand_paths(*sys_path))
        self._site_packages_entries = tuple(self._expand_paths(*site_packages))
        self._extras_paths_entries = tuple(self._expand_paths(*extras_paths))
        self._is_venv = is_venv

    @property
    def is_venv(self):
        # type: () -> bool
        return self._is_venv

    def __contains__(self, entry):
        # type: (str) -> bool
        for path in self._expand_paths(entry):
            # N.B.: Since site packages is contained in sys.path, process rejections 1st to ensure
            # they never sneak by.
            if path.startswith(self._site_packages_entries):
                return False
            if path.startswith(self._extras_paths_entries):
                return False
            if path.startswith(self._sys_path_entries):
                return True

        # A sys.path entry injected by calling code. This can happen when a PEX is used as a
        # sys.path entry via the __pex__ importer mechanism or via the old pex_bootstrapper
        # bootstrap_pex_env API used by at least lambdex and pyuwsgi_pex. Pointedly, the AWS lambda
        # does this, adding /var/runtime to sys.path before executing user code.
        return False


class PEX(object):  # noqa: T000
    """PEX, n.

    A self-contained python environment.
    """

    class Error(Exception):
        pass

    class NotFound(Error):
        pass

    class InvalidEntryPoint(Error):
        pass

    @classmethod
    def _clean_environment(cls, env=None, strip_pex_env=True):
        env = env or os.environ
        if strip_pex_env:
            for key in list(env):
                if key.startswith("PEX_"):
                    del env[key]

    def __init__(
        self,
        pex=sys.argv[0],  # type: str
        interpreter=None,  # type: Optional[PythonInterpreter]
        env=ENV,  # type: Variables
        verify_entry_point=False,  # type: bool
    ):
        # type: (...) -> None
        self._pex = pex
        self._interpreter = interpreter or PythonInterpreter.get()
        self._pex_info = PexInfo.from_pex(self._pex)
        self._pex_info_overrides = PexInfo.from_env(env=env)
        self._vars = env
        self._envs = None  # type: Optional[Iterable[PEXEnvironment]]
        self._activated_dists = None  # type: Optional[Iterable[Distribution]]
        self._layout = None  # type: Optional[Layout.Value]
        if verify_entry_point:
            self._do_entry_point_verification()

    @property
    def layout(self):
        # type: () -> Layout.Value
        if self._layout is None:
            self._layout = Layout.identify_original(self._pex)
        return self._layout

    def pex_info(self, include_env_overrides=True):
        # type: (bool) -> PexInfo
        pex_info = self._pex_info.copy()
        if include_env_overrides:
            pex_info.update(self._pex_info_overrides)
            pex_info.merge_pex_path(self._vars.PEX_PATH)
        return pex_info

    @property
    def interpreter(self):
        # type: () -> PythonInterpreter
        return self._interpreter

    @property
    def _loaded_envs(self):
        # type: () -> Iterable[PEXEnvironment]
        if self._envs is None:
            # set up the local .pex environment
            pex_info = self.pex_info()
            target = LocalInterpreter.create(self._interpreter)
            envs = [PEXEnvironment.mount(self._pex, pex_info, target=target)]
            # N.B. by this point, `pex_info.pex_path` will contain a single pex path
            # merged from pex_path in `PEX-INFO` and `PEX_PATH` set in the environment.
            # `PEX_PATH` entries written into `PEX-INFO` take precedence over those set
            # in the environment.
            for pex_path in pex_info.pex_path:
                # Set up other environments as specified in pex_path.
                pex_info = PexInfo.from_pex(pex_path)
                pex_info.update(self._pex_info_overrides)
                envs.append(PEXEnvironment.mount(pex_path, pex_info, target=target))
            self._envs = tuple(envs)
        return self._envs

    def resolve(self):
        # type: () -> Iterator[Distribution]
        """Resolves all distributions loadable from this PEX by the current interpreter."""
        seen = set()
        for env in self._loaded_envs:
            for dist in env.resolve():
                # N.B.: Since there can be more than one PEX env on the PEX_PATH we take care to
                # de-dup distributions they have in common.
                if dist in seen:
                    continue
                seen.add(dist)
                yield dist

    def _activate(self):
        # type: () -> Iterable[Distribution]

        activated_dists = []  # type: List[Distribution]
        for env in self._loaded_envs:
            activated_dists.extend(env.activate())

        # Ensure that pkg_resources is not imported until at least every pex environment
        # (i.e. PEX_PATH) has been merged into the environment
        PEXEnvironment._declare_namespace_packages(activated_dists)
        return activated_dists

    def activate(self):
        # type: () -> Iterable[Distribution]
        if self._activated_dists is None:
            # 1. Scrub the sys.path to present a minimal Python environment.
            self.patch_sys()
            # 2. Activate all code and distributions in the PEX.
            self._activated_dists = self._activate()
        return self._activated_dists

    @classmethod
    def minimum_sys_modules(
        cls,
        isolated_sys_path,  # type: IsolatedSysPath
        modules=None,  # type: Optional[Mapping[str, ModuleType]]
    ):
        # type: (...) -> Mapping[str, ModuleType]
        """Given a set of site-packages paths, return a "clean" sys.modules.

        When importing site, modules within `sys.modules` have their `__path__`s populated with
        additional paths as defined by `*-nspkg.pth` in `site-packages`, or alternately by
        distribution metadata such as `*.dist-info/namespace_packages.txt`. This can possibly cause
        namespace packages to leak into imports despite being scrubbed from `sys.path`.

        NOTE: This method mutates modules' `__path__` attributes in `sys.modules`, so this is
        currently an irreversible operation.
        """

        is_venv = isolated_sys_path.is_venv
        modules = modules or sys.modules
        new_modules = {}

        for module_name, module in modules.items():
            # Tainted modules should be dropped.
            module_file = getattr(module, "__file__", None)
            if (
                # The `_virtualenv` module is a known special case provided by the virtualenv
                # project. It should not be un-imported or else the virtualenv importer it installs
                # for performing needed patches to the `distutils` stdlib breaks.
                #
                # See:
                # + https://github.com/pantsbuild/pex/issues/992
                # + https://github.com/pypa/virtualenv/pull/1688
                (not is_venv or module_name != "_virtualenv")
                and module_file
                and module_file not in isolated_sys_path
            ):
                TRACER.log("Dropping %s" % (module_name,), V=3)
                continue

            module_path = getattr(module, "__path__", None)
            # Untainted non-packages (builtin modules) need no further special handling and can
            # stay.
            if module_path is None:
                new_modules[module_name] = module
                continue

            # Unexpected objects, e.g. PEP 420 namespace packages, should just be dropped.
            if not isinstance(module_path, list):
                TRACER.log("Dropping %s" % (module_name,), V=3)
                continue

            # Drop tainted package paths.
            for k in reversed(range(len(module_path))):
                if module_path[k] not in isolated_sys_path:
                    TRACER.log("Scrubbing %s.__path__: %s" % (module_name, module_path[k]), V=3)
                    module_path.pop(k)

            # The package still contains untainted path elements, so it can stay.
            if module_path:
                new_modules[module_name] = module

        return new_modules

    _PYTHONPATH = "PYTHONPATH"
    _STASHED_PYTHONPATH = "_PEX_PYTHONPATH"

    @classmethod
    def stash_pythonpath(cls):
        # type: () -> Optional[str]
        pythonpath = os.environ.pop(cls._PYTHONPATH, None)
        if pythonpath is not None:
            os.environ[cls._STASHED_PYTHONPATH] = pythonpath
        return pythonpath

    @classmethod
    def unstash_pythonpath(cls):
        # type: () -> Optional[str]
        pythonpath = os.environ.pop(cls._STASHED_PYTHONPATH, None)
        if pythonpath is not None:
            os.environ[cls._PYTHONPATH] = pythonpath
        return pythonpath

    @classmethod
    def minimum_sys_path(
        cls,
        isolated_sys_path,  # type: IsolatedSysPath
        inherit_path,  # type: InheritPath.Value
    ):
        # type: (...) -> Tuple[List[str], Mapping[str, Any]]
        scrub_paths = OrderedSet()  # type: OrderedSet[str]
        site_distributions = OrderedSet()  # type: OrderedSet[str]
        user_site_distributions = OrderedSet()  # type: OrderedSet[str]

        def all_distribution_paths(path):
            # type: (Optional[str]) -> Iterable[str]
            if path is None:
                return ()
            locations = set(dist.location for dist in find_distributions(path))
            return {path} | locations | set(os.path.realpath(path) for path in locations)

        for path_element in sys.path:
            if path_element not in isolated_sys_path:
                TRACER.log("Tainted path element: %s" % path_element)
                site_distributions.update(all_distribution_paths(path_element))
            else:
                TRACER.log("Not a tainted path element: %s" % path_element, V=2)

        user_site_distributions.update(all_distribution_paths(USER_SITE))

        if inherit_path == InheritPath.FALSE:
            scrub_paths = OrderedSet(site_distributions)
            scrub_paths.update(user_site_distributions)
            for path in user_site_distributions:
                TRACER.log("Scrubbing from user site: %s" % path)
            for path in site_distributions:
                TRACER.log("Scrubbing from site-packages: %s" % path)

        scrubbed_sys_path = list(OrderedSet(sys.path) - scrub_paths)

        pythonpath = cls.unstash_pythonpath()
        if pythonpath is not None:
            original_pythonpath = pythonpath.split(os.pathsep)
            user_pythonpath = list(OrderedSet(original_pythonpath) - set(sys.path))
            if original_pythonpath == user_pythonpath:
                TRACER.log("Unstashed PYTHONPATH of %s" % pythonpath, V=2)
            else:
                TRACER.log(
                    "Extracted user PYTHONPATH of %s from unstashed PYTHONPATH of %s"
                    % (os.pathsep.join(user_pythonpath), pythonpath),
                    V=2,
                )

            if inherit_path == InheritPath.FALSE:
                for path in user_pythonpath:
                    TRACER.log("Scrubbing user PYTHONPATH element: %s" % path)
            elif inherit_path == InheritPath.PREFER:
                TRACER.log("Prepending user PYTHONPATH: %s" % os.pathsep.join(user_pythonpath))
                scrubbed_sys_path = user_pythonpath + scrubbed_sys_path
            elif inherit_path == InheritPath.FALLBACK:
                TRACER.log("Appending user PYTHONPATH: %s" % os.pathsep.join(user_pythonpath))
                scrubbed_sys_path = scrubbed_sys_path + user_pythonpath

        scrub_from_importer_cache = filter(
            lambda key: any(key.startswith(path) for path in scrub_paths),
            sys.path_importer_cache.keys(),
        )
        scrubbed_importer_cache = dict(
            (key, value)
            for (key, value) in sys.path_importer_cache.items()
            if key not in scrub_from_importer_cache
        )

        for importer_cache_entry in scrub_from_importer_cache:
            TRACER.log("Scrubbing from path_importer_cache: %s" % importer_cache_entry, V=2)

        return scrubbed_sys_path, scrubbed_importer_cache

    def minimum_sys(self, inherit_path):
        # type: (InheritPath.Value) -> Tuple[List[str], Mapping[str, Any], Mapping[str, ModuleType]]
        """Return the minimum sys necessary to run this interpreter, a la python -S.

        :returns: (sys.path, sys.path_importer_cache, sys.modules) tuple of a
          bare python installation.
        """
        isolated_sys_path = IsolatedSysPath.for_pex(
            interpreter=self._interpreter, pex=self._pex, pex_pex=self._vars.PEX
        )
        sys_path, sys_path_importer_cache = self.minimum_sys_path(isolated_sys_path, inherit_path)
        sys_modules = self.minimum_sys_modules(isolated_sys_path)

        return sys_path, sys_path_importer_cache, sys_modules

    # Thar be dragons -- when this function exits, the interpreter is potentially in a wonky state
    # since the patches here (minimum_sys_modules for example) actually mutate global state.
    def patch_sys(self):
        # type: () -> None
        """Patch sys with all site scrubbed."""
        inherit_path = self._vars.PEX_INHERIT_PATH
        if inherit_path == InheritPath.FALSE:
            inherit_path = self._pex_info.inherit_path

        def patch_dict(old_value, new_value):
            # type: (Dict[_K, _V], Mapping[_K, _V]) -> None
            old_value.clear()
            old_value.update(new_value)

        def patch_all(path, path_importer_cache, modules):
            # type: (List[str], Mapping[str, Any], Mapping[str, ModuleType]) -> None
            sys.path[:] = path
            patch_dict(sys.path_importer_cache, path_importer_cache)
            patch_dict(sys.modules, modules)

            # N.B.: These hooks may contain custom code set via site.py or `.pth` files that we've
            # just scrubbed. Reset them to factory defaults to avoid scrubbed code.
            sys.displayhook = sys.__displayhook__
            sys.excepthook = sys.__excepthook__  # type: ignore[assignment] # Builtin typeshed bug.

        new_sys_path, new_sys_path_importer_cache, new_sys_modules = self.minimum_sys(inherit_path)

        if self._vars.PEX_EXTRA_SYS_PATH:
            TRACER.log(
                "Adding {} to sys.path".format(os.pathsep.join(self._vars.PEX_EXTRA_SYS_PATH))
            )
            extra_sys_path = self._vars.PEX_EXTRA_SYS_PATH
            new_sys_path.extend(extra_sys_path)

            # Let Python subprocesses see the same sys.path additions we see. If those Python
            # subprocesses are PEX subprocesses, they'll do their own (re-)scrubbing as needed.
            if inherit_path is InheritPath.FALSE:
                pythonpath_entries = extra_sys_path
            else:
                raw_pythonpath = os.environ.get(self._PYTHONPATH)
                pythonpath = tuple(raw_pythonpath.split(os.pathsep)) if raw_pythonpath else ()
                pythonpath_entries = pythonpath + extra_sys_path
            os.environ[self._PYTHONPATH] = os.pathsep.join(pythonpath_entries)

        TRACER.log("New sys.path: {}".format(new_sys_path))

        patch_all(new_sys_path, new_sys_path_importer_cache, new_sys_modules)

    def _wrap_coverage(self, runner, *args):
        if not self._vars.PEX_COVERAGE and self._vars.PEX_COVERAGE_FILENAME is None:
            return runner(*args)

        try:
            import coverage
        except ImportError:
            die("Could not bootstrap coverage module, aborting.")

        pex_coverage_filename = self._vars.PEX_COVERAGE_FILENAME
        if pex_coverage_filename is not None:
            cov = coverage.coverage(data_file=pex_coverage_filename)
        else:
            cov = coverage.coverage(data_suffix=True)

        TRACER.log("Starting coverage.")
        cov.start()

        try:
            return runner(*args)
        finally:
            TRACER.log("Stopping coverage")
            cov.stop()

            # TODO(wickman) Post-process coverage to elide $PEX_ROOT and make
            # the report more useful/less noisy.  #89
            if pex_coverage_filename:
                cov.save()
            else:
                cov.report(show_missing=False, ignore_errors=True, file=sys.stdout)

    def _wrap_profiling(self, runner, *args):
        if not self._vars.PEX_PROFILE and self._vars.PEX_PROFILE_FILENAME is None:
            return runner(*args)

        pex_profile_filename = self._vars.PEX_PROFILE_FILENAME
        pex_profile_sort = self._vars.PEX_PROFILE_SORT
        try:
            import cProfile as profile
        except ImportError:
            import profile

        profiler = profile.Profile()

        try:
            return profiler.runcall(runner, *args)
        finally:
            if pex_profile_filename is not None:
                profiler.dump_stats(pex_profile_filename)
            else:
                profiler.print_stats(sort=pex_profile_sort)

    def path(self):
        # type: () -> str
        """Return the path this PEX was built at."""
        return self._pex

    def execute(self):
        # type: () -> None
        """Execute the PEX.

        This function makes assumptions that it is the last function called by the interpreter.
        """
        if self._vars.PEX_TOOLS:
            if not self._pex_info.includes_tools:
                die(
                    "The PEX_TOOLS environment variable was set, but this PEX was not built "
                    "with tools (Re-build the PEX file with `pex --include-tools ...`)"
                )

            from pex.tools import main as tools

            sys.exit(tools.main(pex=PEX(sys.argv[0])))

        self.activate()

        pex_file = self._vars.PEX
        if pex_file:
            try:
                from setproctitle import setproctitle  # type: ignore[import]

                setproctitle(
                    "{python} {pex_file} {args}".format(
                        python=sys.executable,
                        pex_file=pex_file,
                        args=" ".join(sys.argv[1:]),
                    )
                )
            except ImportError:
                TRACER.log(
                    "Not setting process title since setproctitle is not available in "
                    "{pex_file}".format(pex_file=pex_file),
                    V=3,
                )

        sys.exit(self._wrap_coverage(self._wrap_profiling, self._execute))

    def _execute(self):
        # type: () -> Any
        force_interpreter = self._vars.PEX_INTERPRETER

        self._clean_environment(strip_pex_env=self._pex_info.strip_pex_env)

        if force_interpreter:
            TRACER.log("PEX_INTERPRETER specified, dropping into interpreter")
            return self.execute_interpreter()

        if not any(
            (
                self._pex_info_overrides.script,
                self._pex_info_overrides.entry_point,
                self._pex_info.script,
                self._pex_info.entry_point,
            )
        ):
            TRACER.log("No entry point specified, dropping into interpreter")
            return self.execute_interpreter()

        if self._pex_info_overrides.script and self._pex_info_overrides.entry_point:
            return "Cannot specify both script and entry_point for a PEX!"

        if self._pex_info.script and self._pex_info.entry_point:
            return "Cannot specify both script and entry_point for a PEX!"

        if self._pex_info_overrides.script:
            return self.execute_script(self._pex_info_overrides.script)
        if self._pex_info_overrides.entry_point:
            return self.execute_entry(
                EntryPoint.parse("run = {}".format(self._pex_info_overrides.entry_point))
            )

        for name, value in self._pex_info.inject_env.items():
            os.environ.setdefault(name, value)
        sys.argv[1:1] = list(self._pex_info.inject_args)

        if self._pex_info.script:
            return self.execute_script(self._pex_info.script)
        else:
            return self.execute_entry(
                EntryPoint.parse("run = {}".format(self._pex_info.entry_point))
            )

    @classmethod
    def demote_bootstrap(cls):
        TRACER.log("Bootstrap complete, performing final sys.path modifications...")

        should_log = {level: TRACER.should_log(V=level) for level in range(1, 10)}

        def log(msg, V=1):
            if should_log.get(V, False):
                print("pex: {}".format(msg), file=sys.stderr)

        # Remove the third party resources pex uses and demote pex bootstrap code to the end of
        # sys.path for the duration of the run to allow conflicting versions supplied by user
        # dependencies to win during the course of the execution of user code.
        third_party.uninstall()

        bootstrap = Bootstrap.locate()
        log("Demoting code from %s" % bootstrap, V=2)
        for module in bootstrap.demote():
            log("un-imported {}".format(module), V=9)

        import pex

        log("Re-imported pex from {}".format(pex.__path__), V=3)

        log("PYTHONPATH contains:")
        for element in sys.path:
            log("  %c %s" % (" " if os.path.exists(element) else "*", element))
        log("  * - paths that do not exist or will be imported via zipimport")

    def execute_interpreter(self):
        # type: () -> Any

        # A Python interpreter always inserts the CWD at the head of the sys.path.
        # See https://docs.python.org/3/library/sys.html#sys.path
        sys.path.insert(0, "")

        args = sys.argv[1:]
        python_options = []
        for index, arg in enumerate(args):
            # Check if the arg is an expected startup arg.
            if arg.startswith("-") and arg not in ("-", "-c", "-m"):
                python_options.append(arg)
            else:
                args = args[index:]
                break

        # The pex was called with Python interpreter options
        if python_options:
            return self.execute_with_options(python_options, args)

        if args:
            # NB: We take care here to setup sys.argv to match how CPython does it for each case.
            arg = args[0]
            if arg == "-c":
                content = args[1]
                sys.argv = ["-c"] + args[2:]
                return self.execute_content("-c <cmd>", content, argv0="-c")
            elif arg == "-m":
                module = args[1]
                sys.argv = args[1:]
                return self.execute_module(module)
            else:
                try:
                    if arg == "-":
                        content = sys.stdin.read()
                    else:
                        file_path = arg if os.path.isfile(arg) else os.path.join(arg, "__main__.py")
                        with open(file_path) as fp:
                            content = fp.read()
                except IOError as e:
                    return "Could not open {} in the environment [{}]: {}".format(
                        arg, sys.argv[0], e
                    )
                sys.argv = args
                return self.execute_content(arg, content)
        else:
            if self._vars.PEX_INTERPRETER_HISTORY:
                import atexit
                import readline

                histfile = os.path.expanduser(self._vars.PEX_INTERPRETER_HISTORY_FILE)
                try:
                    readline.read_history_file(histfile)
                    readline.set_history_length(1000)
                except OSError as e:
                    sys.stderr.write(
                        "Failed to read history file at {} due to: {}".format(histfile, e)
                    )

                atexit.register(readline.write_history_file, histfile)

            self.demote_bootstrap()

            import code

            code.interact()
            return None

    def execute_with_options(
        self,
        python_options,  # type: List[str]
        args,  # List[str]
    ):
        # type: (...) -> Union[NoReturn, Any]
        """
        Restart the process passing the given options to the python interpreter
        """
        # Find the installed (unzipped) PEX entry point.
        main = sys.modules.get("__main__")
        if not main or not main.__file__:
            # N.B.: This should never happen.
            return "Unable to resolve PEX __main__ module file: {}".format(main)

        python = sys.executable
        cmdline = [python] + python_options + [main.__file__] + args
        TRACER.log(
            "Re-executing with Python interpreter options: cmdline={cmdline!r}".format(
                cmdline=" ".join(cmdline)
            )
        )
        os.execv(python, cmdline)

    def execute_script(self, script_name):
        # type: (str) -> Any
        dists = list(self.activate())

        dist_entry_point = get_entry_point_from_console_script(script_name, dists)
        if dist_entry_point:
            TRACER.log(
                "Found console_script {!r} in {!r}.".format(
                    dist_entry_point.entry_point, dist_entry_point.dist
                )
            )
            return self.execute_entry(dist_entry_point.entry_point)

        dist_script = get_script_from_distributions(script_name, dists)
        if not dist_script:
            return "Could not find script {!r} in pex!".format(script_name)

        TRACER.log("Found script {!r} in {!r}.".format(script_name, dist_script.dist))
        ast = dist_script.python_script()
        if ast:
            return self.execute_ast(dist_script.path, ast, argv0=script_name)
        else:
            return self.execute_external(dist_script.path)

    @staticmethod
    def execute_external(binary):
        # type: (str) -> Any
        args = [binary] + sys.argv[1:]
        try:
            return Executor.open_process(args).wait()
        except Executor.ExecutionError as e:
            return "Could not invoke script {}: {}".format(binary, e)

    @classmethod
    def execute_content(
        cls,
        name,  # type: str
        content,  # type: str
        argv0=None,  # type: Optional[str]
    ):
        # type: (...) -> Optional[str]
        try:
            program = compile(content, name, "exec", flags=0, dont_inherit=1)
        except SyntaxError as e:
            return "Unable to parse {}: {}".format(name, e)
        return cls.execute_ast(name, program, argv0=argv0)

    @classmethod
    def execute_ast(
        cls,
        name,  # type: str
        program,  # type: ast.AST
        argv0=None,  # type: Optional[str]
    ):
        # type: (...) -> Optional[str]
        cls.demote_bootstrap()

        from pex.compatibility import exec_function

        sys.argv[0] = argv0 or name
        globals_map = globals().copy()
        globals_map["__name__"] = "__main__"
        globals_map["__file__"] = name
        exec_function(program, globals_map)
        return None

    def execute_entry(self, entry_point):
        # type: (EntryPoint) -> Any
        if isinstance(entry_point, CallableEntryPoint):
            return self.execute_entry_point(entry_point)

        return self.execute_module(entry_point.module)

    def execute_module(self, module_name):
        # type: (str) -> None
        self.demote_bootstrap()

        import runpy

        runpy.run_module(module_name, run_name="__main__", alter_sys=True)

    @classmethod
    def execute_entry_point(cls, entry_point):
        # type: (CallableEntryPoint) -> Any
        cls.demote_bootstrap()

        runner = entry_point.resolve()
        return runner()

    def cmdline(self, args=()):
        """The commandline to run this environment.

        :keyword args: Additional arguments to be passed to the application being invoked by the
          environment.
        """
        cmd, _ = self._interpreter.create_isolated_cmd([self._pex] + list(args))
        return cmd

    def run(self, args=(), with_chroot=False, blocking=True, setsid=False, env=None, **kwargs):
        """Run the PythonEnvironment in an interpreter in a subprocess.

        :keyword args: Additional arguments to be passed to the application being invoked by the
          environment.
        :keyword with_chroot: Run with cwd set to the environment's working directory.
        :keyword blocking: If true, return the return code of the subprocess.
          If false, return the Popen object of the invoked subprocess.
        :keyword setsid: If true, run the PEX in a separate operating system session.
        :keyword env: An optional environment dict to use as the PEX subprocess environment. If none is
                      passed, the ambient environment is inherited.
        Remaining keyword arguments are passed directly to subprocess.Popen.
        """
        if env is not None:
            # If explicit env vars are passed, we don't want to clean any of these.
            env = env.copy()
        else:
            env = os.environ.copy()
            self._clean_environment(env=env)

        TRACER.log("PEX.run invoking {}".format(" ".join(self.cmdline(args))))
        _, process = self._interpreter.open_process(
            [self._pex] + list(args),
            cwd=self._pex if with_chroot else os.getcwd(),
            preexec_fn=os.setsid if setsid else None,
            stdin=kwargs.pop("stdin", None),
            stdout=kwargs.pop("stdout", None),
            stderr=kwargs.pop("stderr", None),
            env=env,
            **kwargs
        )
        return process.wait() if blocking else process

    def _do_entry_point_verification(self):

        entry_point = self._pex_info.entry_point
        ep_split = entry_point.split(":")

        # a.b.c:m ->
        # ep_module = 'a.b.c'
        # ep_method = 'm'

        # Only module is specified
        if len(ep_split) == 1:
            ep_module = ep_split[0]
            import_statement = "import {}".format(ep_module)
        elif len(ep_split) == 2:
            ep_module = ep_split[0]
            ep_method = ep_split[1]
            import_statement = "from {} import {}".format(ep_module, ep_method)
        else:
            raise self.InvalidEntryPoint("Failed to parse: `{}`".format(entry_point))

        with named_temporary_file() as fp:
            fp.write(import_statement.encode("utf-8"))
            fp.close()
            retcode = self.run([fp.name], env={"PEX_INTERPRETER": "1"})
            if retcode != 0:
                raise self.InvalidEntryPoint(
                    "Invalid entry point: `{}`\n"
                    "Entry point verification failed: `{}`".format(entry_point, import_statement)
                )



# coding=utf-8
# Copyright 2019 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import warnings

import pytest

from pex import pex_warnings
from pex.compatibility import PY2
from pex.pex_info import PexInfo
from pex.pex_warnings import PEXWarning
from pex.typing import TYPE_CHECKING
from pex.variables import Variables

if TYPE_CHECKING:
    from typing import List


def exercise_warnings(pex_info, **env):
    # type: (PexInfo, **str) -> List[warnings.WarningMessage]
    with warnings.catch_warnings(record=True) as events:
        pex_warnings.configure_warnings(env=Variables(environ=env), pex_info=pex_info)
        pex_warnings.warn("test")
    assert events is not None
    return events


def assert_warnings(pex_info, **env):
    # type: (PexInfo, **str) -> None
    events = exercise_warnings(pex_info, **env)
    assert 1 == len(events)
    warning = events[0]
    assert PEXWarning == warning.category
    assert "test" == str(warning.message)


def assert_no_warnings(pex_info, **env):
    # type: (PexInfo, **str) -> None
    events = exercise_warnings(pex_info, **env)
    assert 0 == len(events)


def pex_info_no_emit_warnings():
    # type: () -> PexInfo
    pex_info = PexInfo.default()
    pex_info.emit_warnings = False
    return pex_info


skip_py2_warnings = pytest.mark.skipif(
    PY2,
    reason=(
        "The `warnings.catch_warnings` mechanism doesn't work properly under CPython 2.7 & pypy2 "
        "across multiple tests. Since we only use `warnings.catch_warnings` in unit tests and "
        "the mechanisms tested here are also tested in integration tests under CPython 2.7 & pypy "
        "we accept that these unit tests appear un-fixable without a lot of warnings mocking."
    ),
)


@skip_py2_warnings
def test_emit_warnings_default_on():
    # type: () -> None
    assert_warnings(PexInfo.default())


@skip_py2_warnings
def test_emit_warnings_pex_info_off():
    # type: () -> None
    assert_no_warnings(pex_info_no_emit_warnings())


@skip_py2_warnings
def test_emit_warnings_emit_env_off():
    # type: () -> None
    assert_no_warnings(PexInfo.default(), PEX_EMIT_WARNINGS="0")


@skip_py2_warnings
def test_emit_warnings_pex_info_off_emit_env_override():
    # type: () -> None
    assert_warnings(pex_info_no_emit_warnings(), PEX_EMIT_WARNINGS="1")


@skip_py2_warnings
def test_emit_warnings_pex_info_off_verbose_override():
    # type: () -> None
    assert_warnings(pex_info_no_emit_warnings(), PEX_VERBOSE="1")


@skip_py2_warnings
def test_emit_warnings_pex_info_off_verbose_trumps_emit_env():
    # type: () -> None
    assert_warnings(pex_info_no_emit_warnings(), PEX_VERBOSE="1", PEX_EMIT_WARNINGS="0")



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import errno
import json
import os
import subprocess
from textwrap import dedent
from typing import Any, Callable, List

import colors
import pytest

from pex.cli.testing import run_pex3
from pex.common import chmod_plus_x, touch
from pex.testing import IS_PYPY, run_pex_command
from pex.version import __version__


def find_max_length(
    seed_max,  # type: int
    is_too_long,  # type: Callable[[int], bool]
):
    # type: (...) -> int

    too_long_low_watermark = seed_max
    ok_high_watermark = 0
    current_length = seed_max
    steps = 0
    while True:
        steps += 1
        if is_too_long(current_length):
            too_long_low_watermark = min(too_long_low_watermark, current_length)
        elif current_length + 1 == too_long_low_watermark:
            return current_length
        else:
            assert current_length < seed_max, "Did not probe high enough for shebang length limit."
            ok_high_watermark = max(ok_high_watermark, current_length)
        assert ok_high_watermark < too_long_low_watermark
        current_length = ok_high_watermark + (too_long_low_watermark - ok_high_watermark) // 2


# Pytest fails to cleanup tmp dirs used probing file_path_length_limit and this squashes a very
# large ream of warnings.
pytestmark = pytest.mark.filterwarnings("ignore:\\(rm_rf\\) error removing.*:pytest.PytestWarning")


@pytest.fixture(scope="module")
def file_path_length_limit(tmpdir_factory):
    # type: (Any) -> int

    def file_path_too_long(length):
        # type: (int) -> bool
        path = str(tmpdir_factory.mktemp("td"))
        while len(path) < length - len(os.path.join("directory", "x")):
            path = os.path.join(path, "directory")
            try:
                os.mkdir(path)
            except (IOError, OSError) as e:
                if e.errno == errno.ENAMETOOLONG:
                    return True
                elif e.errno != errno.EEXIST:
                    raise e

        if len(path) < length:
            padding = length - len(path) - len(os.sep)
            path = os.path.join(path, "x" * padding)
            try:
                touch(path)
            except (IOError, OSError) as e:
                if e.errno == errno.ENAMETOOLONG:
                    return True
                raise e

        return False

    return find_max_length(seed_max=2 ** 16, is_too_long=file_path_too_long)


@pytest.fixture(scope="module")
def shebang_length_limit(
    tmpdir_factory,  # type: Any
    file_path_length_limit,  # type: int
):
    # type: (...) -> int

    def shebang_too_long(length):
        # type: (int) -> bool
        path = str(tmpdir_factory.mktemp("td"))
        while len(path) < length - len("#!\n" + os.path.join("directory", "x")):
            path = os.path.join(path, "directory")
            try:
                os.mkdir(path)
            except (IOError, OSError) as e:
                if e.errno != errno.EEXIST:
                    raise e

        sh_path = os.path.join(path, "x" * (length - len("#!\n" + path + os.sep)))
        try:
            os.unlink(sh_path)
        except (IOError, OSError) as e:
            if e.errno != errno.ENOENT:
                raise e
        os.symlink("/bin/sh", sh_path)

        script = os.path.join(path, "script.sh")
        with open(script, "w") as fp:
            fp.write("#!{sh_path}\n".format(sh_path=sh_path))
            fp.write("exit 0\n")
        chmod_plus_x(script)
        try:
            return 0 != subprocess.call(args=[script])
        except (IOError, OSError) as e:
            if e.errno == errno.ENOEXEC:
                return True
            raise e

    return find_max_length(
        seed_max=file_path_length_limit - len(os.sep + "script.sh"), is_too_long=shebang_too_long
    )


@pytest.fixture
def too_deep_pex_root(
    tmpdir,  # type: Any
    file_path_length_limit,  # type: int
    shebang_length_limit,  # type: int
):
    # type: (...) -> str

    # The short venv python used in --venv shebangs is of the form:
    #   <PEX_ROOT>/venvs/s/592c68dc/venv/bin/python
    # With no collisions, the hash dir is 8 characters, and we expect no collisions in this bespoke
    # new empty temporary dir PEX_ROOT>
    padding_dirs_length = shebang_length_limit - len(
        "#!"
        + os.path.join(
            str(tmpdir),
            "pex_root",
            "venvs",
            "s",
            "12345678",
            "venv",
            "bin",
            "pypy" if IS_PYPY else "python",
        )
        + "\n"
    )
    if padding_dirs_length > file_path_length_limit:
        pytest.skip(
            "Cannot create a PEX_ROOT in the tmp dir that both generates a too-long venv pex "
            "script shebang and yet does not generate a path to that venv pex script that is too "
            "long.\n"
            "Max shebang length: {shebang_length_limit}\n"
            "Max file path length: {file_path_length_limit}\n"
            "Temp dir length: {tmpdir_path_length}\n"
            "Temp dir:\n{tmpdir}".format(
                shebang_length_limit=shebang_length_limit,
                file_path_length_limit=file_path_length_limit,
                tmpdir_path_length=len(str(tmpdir)),
                tmpdir=tmpdir,
            )
        )

    padding_dirs_path = "directory"
    while len(padding_dirs_path) < padding_dirs_length - len(os.path.join("directory", "x")):
        padding_dirs_path = os.path.join(padding_dirs_path, "directory")
    padding_dirs_path = os.path.join(
        padding_dirs_path, "x" * (padding_dirs_length - len(padding_dirs_path + os.sep))
    )

    return os.path.realpath(os.path.join(str(tmpdir), padding_dirs_path, "pex_root"))


@pytest.mark.parametrize(
    "execution_mode_args",
    [
        pytest.param([], id="ZIPAPP"),
        pytest.param(["--venv"], id="VENV"),
        pytest.param(["--sh-boot"], id="ZIPAPP (--sh-boot)"),
        pytest.param(["--venv", "--sh-boot"], id="VENV (--sh-boot)"),
    ],
)
def test_shebang_length_limit_runtime(
    tmpdir,  # type: Any
    execution_mode_args,  # type: List[str]
    too_deep_pex_root,  # type: str
):
    # type: (...) -> None

    pex = os.path.realpath(os.path.join(str(tmpdir), "pex"))
    result = run_pex_command(
        args=[
            "--pex-root",
            too_deep_pex_root,
            "--runtime-pex-root",
            too_deep_pex_root,
            "-o",
            pex,
            "--seed",
            "verbose",
        ]
        + execution_mode_args
    )
    result.assert_success()
    seeded_pex = json.loads(result.output)["pex"]

    test_pex_args = ["-c", "import __main__; print(__main__.__file__)"]

    def assert_pex_works(pex_file):
        assert (
            subprocess.check_output(args=[pex_file] + test_pex_args)
            .decode("utf8")
            .startswith(too_deep_pex_root)
        )

    if "--venv" in execution_mode_args:
        # Running the venv pex directly should fail since the shebang length is too long.
        with pytest.raises(OSError) as exc_info:
            subprocess.check_call(args=[seeded_pex] + test_pex_args)
        assert exc_info.value.errno == errno.ENOEXEC
    else:
        assert_pex_works(seeded_pex)

    assert_pex_works(pex)


def test_shebang_length_limit_buildtime_resolve(
    tmpdir,  # type: Any
    too_deep_pex_root,  # type: str
):
    # type: (...) -> None

    pex = os.path.realpath(os.path.join(str(tmpdir), "pex"))
    # N.B.: This runs the vendored Pip tool in a dogfood venv to resolve ansicolors.
    run_pex_command(
        args=[
            "--pex-root",
            too_deep_pex_root,
            "--runtime-pex-root",
            too_deep_pex_root,
            "-o",
            pex,
            "ansicolors==1.1.8",
        ]
    ).assert_success()

    assert (
        colors.cyan("Jane")
        == subprocess.check_output(args=[pex, "-c", "import colors; print(colors.cyan('Jane'))"])
        .decode("utf-8")
        .strip()
    )


def test_shebang_length_limit_buildtime_lock_local_project(
    tmpdir,  # type: Any
    pex_project_dir,  # type: str
    too_deep_pex_root,  # type: str
):
    # type: (...) -> None

    lock = os.path.realpath(os.path.join(str(tmpdir), "lock.json"))
    # N.B.: This runs the vendored PEP 517 / 518 sdist build tool in a dogfood venv to create an
    # sdist for the local Pex project that can be consistently hashed.
    run_pex3(
        "lock",
        "create",
        "--pex-root",
        too_deep_pex_root,
        "-o",
        lock,
        "--indent",
        "2",
        "ansicolors==1.1.8",
        pex_project_dir,
    ).assert_success()

    pex = os.path.realpath(os.path.join(str(tmpdir), "pex"))
    run_pex_command(
        args=[
            "--pex-root",
            too_deep_pex_root,
            "--runtime-pex-root",
            too_deep_pex_root,
            "-o",
            pex,
            "--lock",
            lock,
        ]
    ).assert_success()

    assert (
        colors.yellow(__version__)
        == subprocess.check_output(
            args=[
                pex,
                "-c",
                dedent(
                    """\
                    import colors

                    from pex.version import __version__


                    print(colors.yellow(__version__))
                    """
                ),
            ]
        )
        .decode("utf-8")
        .strip()
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).
import hashlib
import itertools
from collections import defaultdict

import pytest

from pex.resolve.pep_691.model import File
from pex.resolve.resolved_requirement import Fingerprint
from pex.sorted_tuple import SortedTuple
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import DefaultDict, List, Tuple


def file(**hashes):
    # type (**str) -> File
    return File(
        filename="foo.tar.gz",
        url="https://files.example.org/foo.tar.gz",
        hashes=SortedTuple(Fingerprint(algorithm, hash_) for algorithm, hash_ in hashes.items()),
    )


def test_select_fingerprint_none():
    # type: () -> None

    assert file().select_fingerprint() is None


@pytest.mark.parametrize("guaranteed_algorithm", sorted(hashlib.algorithms_guaranteed))
def test_select_fingerprint_single_guaranteed(guaranteed_algorithm):
    # type: (str) -> None

    hashes = {guaranteed_algorithm: "hash_hex_value"}
    assert (
        Fingerprint(algorithm=guaranteed_algorithm, hash="hash_hex_value")
        == file(**hashes).select_fingerprint()
    )


def test_select_fingerprint_sha256_trumps_all_others():
    # type: () -> None

    hashes = {algorithm: "hash_hex_value" for algorithm in hashlib.algorithms_available}
    hashes["sha256"] = "preferred"
    assert Fingerprint(algorithm="sha256", hash="preferred") == file(**hashes).select_fingerprint()


def test_select_fingerpint_none_guaranteed():
    # type: () -> None

    hashes = {
        algorithm: "hash_hex_value"
        for algorithm in set(hashlib.algorithms_available) - set(hashlib.algorithms_guaranteed)
    }
    assert file(**hashes).select_fingerprint() is None


@pytest.mark.parametrize(
    "guaranteed_algorithm_pair",
    [
        pytest.param(
            (one, two),
            id="{one}[{one_size}] vs {two}[{two_size}]".format(
                one=one,
                one_size=hashlib.new(one).digest_size,
                two=two,
                two_size=hashlib.new(two).digest_size,
            ),
        )
        for one, two in sorted(
            itertools.combinations(set(hashlib.algorithms_guaranteed) - {"sha256"}, 2)
        )
    ],
)
def test_select_fingerprint_ranking(guaranteed_algorithm_pair):
    # type: (Tuple[str, str]) -> None

    algorithms_by_size = defaultdict(list)  # type: DefaultDict[int, List[str]]
    hashes = {}
    for algorithm in guaranteed_algorithm_pair:
        algorithms_by_size[hashlib.new(algorithm).digest_size].append(algorithm)
        hashes[algorithm] = "{algorithm}_hash_hex_value".format(algorithm=algorithm)

    # We expect algorithms are preferred for greatest digest size and then tie-broken by alphabetic
    # order.
    expected_algorithm = sorted(algorithms_by_size[max(algorithms_by_size)])[0]
    expected_hash = hashes[expected_algorithm]

    assert (
        Fingerprint(algorithm=expected_algorithm, hash=expected_hash)
        == file(**hashes).select_fingerprint()
    )



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
from textwrap import dedent

from pex.common import safe_open, temporary_dir
from pex.testing import PY27, PY310, ensure_python_interpreter, make_env, run_pex_command


def test_top_level_requirements_requires_python_env_markers():
    # type: () -> None
    python27 = ensure_python_interpreter(PY27)
    python310 = ensure_python_interpreter(PY310)
    with temporary_dir() as td:
        src_dir = os.path.join(td, "src")
        with safe_open(os.path.join(src_dir, "test_issues_898.py"), "w") as fp:
            fp.write(
                dedent(
                    """
                    import zipp

                    print(zipp.__file__)
                    """
                )
            )

        pex_file = os.path.join(td, "zipp.pex")

        results = run_pex_command(
            args=[
                "--python={}".format(python27),
                "--python={}".format(python310),
                "zipp>=1,<=3.1.0",
                "--sources-directory={}".format(src_dir),
                "--entry-point=test_issues_898",
                "-o",
                pex_file,
            ],
        )
        results.assert_success()

        pex_root = os.path.realpath(os.path.join(td, "pex_root"))
        for python in python27, python310:
            output = subprocess.check_output([python, pex_file], env=make_env(PEX_ROOT=pex_root))
            zipp_location = os.path.realpath(output.decode("utf-8").strip())
            assert zipp_location.startswith(
                pex_root
            ), "Failed to import zipp from {} under {}".format(pex_file, python)



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import filecmp
import itertools
import json
import os
import signal
import subprocess
from textwrap import dedent

import pytest

from pex.common import DETERMINISTIC_DATETIME, open_zip, safe_open, temporary_dir
from pex.dist_metadata import Distribution, Requirement
from pex.testing import PY310, ensure_python_venv, run_command_with_jitter, run_pex_command
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict, Iterator


@pytest.fixture(scope="module")
def pex():
    # type: () -> Iterator[str]
    with temporary_dir() as tmpdir:
        pex_path = os.path.join(tmpdir, "example.pex")

        src = os.path.join(tmpdir, "src")
        with safe_open(os.path.join(src, "data", "url.txt"), "w") as fp:
            fp.write("https://example.com")
        with safe_open(os.path.join(src, "main.py"), "w") as fp:
            fp.write(
                dedent(
                    """\
                    from __future__ import print_function

                    import os
                    import sys

                    import requests


                    def do():
                        with open(os.path.join(os.path.dirname(__file__), "data", "url.txt")) as fp:
                            url = fp.read().strip()
                        print("Fetching from {} ...".format(url))
                        print(requests.get(url).text, file=sys.stderr)
                    """
                )
            )

        constraints = os.path.join(str(tmpdir), "constraints.txt")
        with open(constraints, "w") as fp:
            # N.B.: urllib3 1.26.10 dropped support for Python 3.5 which we test against.
            fp.write("urllib3<1.26.10")

        result = run_pex_command(
            args=[
                "-D",
                src,
                "requests==2.25.1",
                "--constraints",
                constraints,
                "-e",
                "main:do",
                "-o",
                pex_path,
                "--include-tools",
            ],
        )
        result.assert_success()
        yield os.path.realpath(pex_path)


@pytest.fixture(scope="module")
def pex_tools_env():
    # type: () -> Dict[str, str]
    env = os.environ.copy()
    env.update(PEX_TOOLS="1")
    return env


def test_info(pex, pex_tools_env):
    # type: (str, Dict[str, str]) -> None
    output = subprocess.check_output(args=[pex, "repository", "info"], env=pex_tools_env)
    distributions = {}
    for line in output.decode("utf-8").splitlines():
        name, version, location = str(line).split(" ", 2)
        distribution = Distribution.load(location)
        assert name == distribution.project_name
        assert version == distribution.version
        distributions[name] = version

    assert {"certifi", "chardet", "idna", "requests", "urllib3"} == set(distributions.keys())
    assert "2.25.1" == distributions["requests"]


def test_info_verbose(pex, pex_tools_env):
    # type: (str, Dict[str, str]) -> None
    output = subprocess.check_output(args=[pex, "repository", "info", "-v"], env=pex_tools_env)
    infos = {}
    for line in output.decode("utf-8").splitlines():
        info = json.loads(line)
        distribution = Distribution.load(info["location"])
        project_name = info["project_name"]
        assert distribution.project_name == project_name
        assert distribution.version == info["version"]
        infos[project_name] = info

    assert {"certifi", "chardet", "idna", "requests", "urllib3"} == set(infos.keys())

    requests_info = infos["requests"]
    assert "2.25.1" == requests_info["version"]
    assert SpecifierSet("!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,>=2.7") == SpecifierSet(
        requests_info["requires_python"]
    )
    assert {
        Requirement.parse(req)
        for req in (
            'PySocks!=1.5.7,>=1.5.6; extra == "socks"',
            "certifi>=2017.4.17",
            "chardet<5,>=3.0.2",
            'cryptography>=1.3.4; extra == "security"',
            "idna<3,>=2.5",
            'pyOpenSSL>=0.14; extra == "security"',
            "urllib3<1.27,>=1.21.1",
            'win-inet-pton; (sys_platform == "win32" and python_version == "2.7") and extra == "socks"',
        )
    } == {Requirement.parse(req) for req in requests_info["requires_dists"]}


def test_extract_deterministic_timestamp(pex, pex_tools_env, tmpdir):
    deterministic_date_time = (
        DETERMINISTIC_DATETIME.year,
        DETERMINISTIC_DATETIME.month,
        DETERMINISTIC_DATETIME.day,
        DETERMINISTIC_DATETIME.hour,
        DETERMINISTIC_DATETIME.minute,
        DETERMINISTIC_DATETIME.second,
    )

    deterministic_dists_dir = os.path.join(str(tmpdir), "deterministic-dists")
    subprocess.check_call(
        args=[pex, "repository", "extract", "-f", deterministic_dists_dir], env=pex_tools_env
    )
    with open_zip(
        os.path.join(deterministic_dists_dir, "requests-2.25.1-py2.py3-none-any.whl")
    ) as zipfile:
        infolist = zipfile.infolist()
        assert len(infolist) > 0
        for info in infolist:
            assert deterministic_date_time == info.date_time

    non_deterministic_dists_dir = os.path.join(str(tmpdir), "non_deterministic_dists_dir")
    subprocess.check_call(
        args=[pex, "repository", "extract", "-f", non_deterministic_dists_dir, "--use-system-time"],
        env=pex_tools_env,
    )
    with open_zip(
        os.path.join(non_deterministic_dists_dir, "requests-2.25.1-py2.py3-none-any.whl")
    ) as zipfile:
        infolist = zipfile.infolist()
        assert len(infolist) > 0
        for info in infolist:
            assert deterministic_date_time != info.date_time


def test_extract_non_deterministic_wheels(pex, pex_tools_env):
    dist_dirs = run_command_with_jitter(
        args=[pex, "repository", "extract", "--use-system-time"],
        path_argument="-f",
        extra_env=pex_tools_env,
        count=3,
    )
    for dists_dir1, dists_dir2 in itertools.combinations(dist_dirs, 2):
        dists1 = sorted(os.listdir(dists_dir1))
        assert len(dists1) > 1
        assert dists1 == sorted(os.listdir(dists_dir2))

        same, different, non_reg = filecmp.cmpfiles(
            dists_dir1, dists_dir2, common=dists1, shallow=False
        )
        assert not same
        assert not non_reg
        assert sorted(dists1) == sorted(different)


def test_extract_deterministic_wheels(pex, pex_tools_env):
    # We already have tests that ensure PEX file creation is deterministic; so here we just test
    # that extracting wheels from a fixed PEX file is also deterministic.

    dist_dirs = run_command_with_jitter(
        args=[pex, "repository", "extract"], path_argument="-f", extra_env=pex_tools_env, count=3
    )
    dists_dir1 = dist_dirs.pop()
    dists1 = sorted(os.listdir(dists_dir1))
    assert len(dists1) > 1
    for dists_dir2 in dist_dirs:
        assert dists1 == sorted(os.listdir(dists_dir2))
        same, different, non_reg = filecmp.cmpfiles(
            dists_dir1, dists_dir2, common=dists1, shallow=False
        )
        assert not different
        assert not non_reg
        assert sorted(dists1) == sorted(same)


def test_extract_lifecycle(pex, pex_tools_env, tmpdir):
    # type: (str, Dict[str, str], Any) -> None
    dists_dir = os.path.join(str(tmpdir), "dists")
    pid_file = os.path.join(str(tmpdir), "pid")
    os.mkfifo(pid_file)
    find_links_server = subprocess.Popen(
        args=[
            pex,
            "repository",
            "extract",
            "--serve",
            "--sources",
            "--dest-dir",
            dists_dir,
            "--pid-file",
            pid_file,
        ],
        env=pex_tools_env,
        stdout=subprocess.PIPE,
    )
    with open(pid_file) as fp:
        _, port = fp.read().strip().split(":", 1)
    example_sdist_pex = os.path.join(str(tmpdir), "example-sdist.pex")
    find_links_url = "http://localhost:{}".format(port)
    result = run_pex_command(
        args=[
            "--no-pypi",
            "--find-links",
            find_links_url,
            "example",
            "-c",
            "example",
            "-o",
            example_sdist_pex,
        ]
    )
    result.assert_success()

    _, pip = ensure_python_venv(PY310)
    subprocess.check_call(
        args=[pip, "install", "--no-index", "--find-links", find_links_url, "example"]
    )
    example_console_script = os.path.join(os.path.dirname(pip), "example")

    find_links_server.send_signal(signal.SIGQUIT)
    assert -1 * int(signal.SIGQUIT) == find_links_server.wait()

    expected_output = b"Fetching from https://example.com ...\n"
    assert expected_output == subprocess.check_output(args=[example_sdist_pex])
    assert expected_output == subprocess.check_output(args=[example_console_script])



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from typing import Callable

import pytest

from pex.build_system.testing import assert_build_sdist
from pex.testing import PY_VER
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.skipif(
    PY_VER < (3, 7), reason="This version of Poetry only supports Python 3.7 and greater."
)
def test_build_sdist_pyproject_toml(
    tmpdir,  # type: Any
    clone,  # type: Callable[[str, str], str]
):
    # type: (...) -> None

    # The Poetry backend is important to supprt and the Poetry project dogfoods itself in its build.
    project_dir = clone(
        "https://github.com/python-poetry/poetry",
        "7cd761a6a0bdec7fab76cf742ee4109db854ab76",
    )
    assert_build_sdist(project_dir, "poetry", "1.2.1", tmpdir)


def test_build_sdist_setup_py(
    tmpdir,  # type: Any
    clone,  # type: Callable[[str, str], str]
):
    # type: (...) -> None

    # This is an old setup.py based project that spews interfering output to stdout.
    project_dir = clone(
        "https://github.com/wickman/pystachio", "43acf709464e47ab0f40b26ec3b9dbbdb4e2ef12"
    )
    assert_build_sdist(project_dir, "pystachio", "0.8.10", tmpdir)
